(defpackage :net.acdw.clarence
  (:nicknames #:clarence)
  (:use :cl)
  (:import-from :cl-ppcre
                #:split
                #:regex-replace-all)
  (:import-from #:ppcre
                #:regex-replace-all))

(in-package :net.acdw.clarence)

(defparameter *tags*
  '(("###" . :h3)
    ("##" . :h2)
    ("#" . :h1)
    ("=>" . :link)
    ("> " . :quote)
    ("* " . :list)
    ("```" . :verbatim))
  "Alist of source sigil to internal tags.")

(defparameter *default-line-format* "~&~A"
  "The default format-specifier for lines.")

(defparameter *default-stanza-format* "~&~{~A~^~%~}~%~%"
  "The default format-specifier for stanzas.
Should contain a looping construct.")

(defparameter *default-document-format* "~{~A~&~}"
  "The default format-specifier for documents.
Should contain a looping construct.")

;;; Gemtext documents

;; Basic types

(defstruct tagged
  "A tagged structure."
  (tag nil :type (or null keyword))
  (content nil))

(defstruct (line (:include tagged))
  "A tagged line of gemtext content.")

(defstruct (stanza (:include tagged (content () :type t))
                   (:constructor %make-stanza))
  "A tagged group of lines.
Stanzas are usually groups of all the same tag, but plain stanzas can include
singular links.  A blank line always separates stanzas.
A stanza's tag should be equal to the tag of its first line.")

(defstruct document
  "A group of stanzas representing a complete document.
Note that there is no tag."
  content)

;; Derived types

(defstruct (link (:include line (tag :link :read-only t)))
  "A link to another document or resource, given by HREF."
  (href nil :type (or null string)))

(defstruct (verbatim (:include stanza (tag :verbatim :read-only t))
                     (:constructor %make-verbatim))
  "A verbatim block of text.
This is the only stanza type that might contain a blank line.  ALT is optional
alt-text, and OPEN and CLOSE are used by writers internally."
  (alt nil :type (or null string))
  (open nil :type (or null string))
  (close nil :type (or null string)))

;; Blank lines

(defvar blank (make-line :tag :blank)
  "A blank line.")

(defmethod blankp ((line line))
  "Is LINE blank?"
  (or (tag-equal line :blank)
      (null (line-content line))
      (zerop (length (line-content line)))))

(defmethod blankp ((stanza stanza))
  "Is STANZA blank?"
  (or (null (stanza-content stanza))
      (tag-equal stanza :blank)))

(defmethod blankp ((doc document))
  (null (document-content doc)))

(defmethod blankp ((list list))
  (null list))

;; Tag equality

(defmethod tag-equal ((x null) y) nil)

(defmethod tag-equal (x (y null)) nil)

(defmethod tag-equal ((tag1 symbol) (tag2 symbol))
  (eq tag1 tag2))

(defmethod tag-equal ((tag symbol) (x tagged))
  (eq tag (slot-value x 'tag)))

(defmethod tag-equal ((x tagged) (tag symbol))
  (eq tag (slot-value x 'tag)))

(defmethod tag-equal ((x tagged) (y tagged))
  (eq (slot-value x 'tag) (slot-value y 'tag)))

(defun tag-null (x)
  (tag-equal x nil))

;; Constructors

(defun make-stanza (lines)
  "Make a STANZA from LINES.
If the stanza is verbatim, remove the first and last lines from its
content---those are each \"```\"."
  (if (tag-equal (car lines) :verbatim)
      (%make-verbatim :content (cdr lines)
                      :alt (line-content (car lines)))
      (%make-stanza :tag (line-tag (car lines))
                    :content lines)))

;;; Reading

(defun tag-of (string)
  "Return the tag of STRING by consulting *TAGS*."
  (loop for (sigil . tag) in *tags*
        when (uiop:string-prefix-p sigil string)
          do (return tag)))

(defun parse-line (string &key verbatimp)
  "Parse STRING as a LINE.
If VERBATIMP is non-nil, parse STRING verbatim."
  (let* ((tag (tag-of string))
         (text (chomp (subseq string (length (car (rassoc tag *tags*)))))))
    (cond
      (verbatimp
       (make-line :tag :verbatim :content text))
      ((eq :link tag)
       (let* ((%split-line (split "\\s" text))
              (href (chomp (car %split-line)))
              (text (if (cdr %split-line)
                        (chomp (format nil "~{~A~^ ~}" (cdr %split-line)))
                        href)))
         (make-link :href href :content text)))
      ((zerop (length text))
       blank)
      (t
       (make-line :tag tag :content text)))))

(defun parse-lines (&optional (input *standard-input*))
  "Parse INPUT into a list of LINEs."
  (loop for this = (read-line input nil)
        with verbatimp = nil
        until (null this)
        when (eq :verbatim (tag-of this))
          do (setf verbatimp (not verbatimp))
        collect (parse-line this :verbatimp verbatimp)))

(defun group-stanzas (lines)
  "Group LINES into a list of STANZAs."
  (flet ((%accumulate-stanzas (stanzas line)
           (let ((stanza (car stanzas)))
             (cond
               ;; A naked line goes into a stanza by itself.
               ((not stanza)
                (list (list line) nil))
               ;; Two links in a row -- start a new stanza.
               ((and (tag-equal line :link)
                     (tag-equal (car stanza) :link)
                     (tag-null (cadr stanza)))
                (list* (list line (car stanza))
                       (make-stanza (nreverse (cdr stanza)))
                       (cdr stanzas)))
               ;; Same line type -- continue this stanza.
               ((tag-equal (car stanza) (line-tag line))
                (list* (cons line stanza)
                       (cdr stanzas)))
               ;; Links within paragraphs --- continue the stanza.
               ((or (and (tag-equal line :link)
                         (tag-null (car stanza))
                         (not (blankp (car stanza))))
                    (and (tag-null line)
                         (tag-equal (car stanza) :link)
                         (tag-null (cadr stanza))))
                (list* (cons line stanza)
                       (cdr stanzas)))
               ;; Different tags --- start a new stanza.
               (t
                (list* (list line)
                       (make-stanza (nreverse stanza))
                       (cdr stanzas)))))))
    (let ((stanzas (remove-if #'blankp
                              (reduce #'%accumulate-stanzas lines
                                      :initial-value ()))))
      (remove-if (lambda (x) (or (null x) (blankp x)))
                 (nreverse (cons (make-stanza (nreverse (car stanzas)))
                                 (cdr stanzas)))))))

(defun parse (source)
  (make-document :content (group-stanzas (parse-lines source))))

(defun parse-string (string)
  (with-input-from-string (s string)
    (parse s)))

(defun parse-file (file)
  (with-open-file (s file)
    (parse s)))

;;; Writing

(defstruct writer
  (name nil :type (or symbol string character))
  ;; FORMATS, STRUCT-FILTERS, and TEXT-FILTERS are all lists of SPECs, where a
  ;; SPEC is of the form (TAG LEVEL PAYLOAD).
  (formats `((:line ,*default-line-format*)
             (:stanza ,*default-stanza-format*)
             (:document ,*default-document-format*)))
  struct-filters
  text-filters)

(defparameter *current-writer* (make-writer))
(defparameter *writers* (list *current-writer*))

(defun current-writer ()
  (or *current-writer* (make-writer)))

(defmacro defwriter (name &key formats struct-filters text-filters)
  `(let ((new-writer (make-writer :name ,(intern (string name) :keyword)
                                  :formats ',formats
                                  :struct-filters ',struct-filters
                                  :text-filters ',text-filters)))
     (pushnew new-writer
              *writers*
              :test #'equal)
     new-writer))

(defwriter gemtext
  :formats ((:h3 (:line "~&### ~A"))
            (:h2 (:line "~&## ~A"))
            (:h1 (:line "~&# ~A"))
            (:quote (:line "~&> ~A"))
            (:list (:line "~&* ~A"))
            (:link (:line "~&=> ~A ~A" href content))
            (:verbatim (:line "~&~A"))
            (:verbatim (:stanza  "~&```~A~%~{~A~%~}```~&" alt content)))
  :struct-filters ((:stanza combine-like))
  :text-filters ((:line unfold-text)))

(defwriter html
  :formats ((:stanza "~&<p>~{~A~^~%~}</p>")
            (:verbatim
             (:stanza "~&~A~{~A~^~%~}~A" open content close))
            (:link
             (:line "~&<li><a href=\"~A\">~A</a></li>" href content)
             (:stanza "~&<ul>~%~{~A~%~}</ul>")
             (:inline "<a href=\"~A\">~A</a> " href content))
            (:list (:line "~&<li>~A</li>")
                   (:stanza "~&<ul>~%~{~A~%~}</ul>"))
            (:quote (:stanza "~&<blockquote>~%~{~A~%~}</blockquote>"))
            (:h1 (:stanza "~&<h1>~{~A~^ ~}</h1>"))
            (:h2 (:stanza "~&<h2>~{~A~^ ~}</h2>"))
            (:h3 (:stanza "~&<h3>~{~A~^ ~}</h3>")))
  :struct-filters ((:line add-inline-markup escape-html/pre)
                   (:verbatim (:stanza verbatim-preproc))
                   (:verbatim (:line identity))))

(defun writer (x)
  (etypecase x
    (writer x)
    ((or symbol string character)       ; string-designator
     (or (find x *writers*
               :key #'writer-name
               :test #'string-equal)
         (error "Writer `~A' unknown." x)))))

(defun spec-match (formats tag level)
  (if (levelp tag)
      (alist-get tag formats)
      (alist-get level (alist-get tag formats))))

(defun levelp (x)
  (find x '(:line :inline :stanza :document)))

(defun format-object (writer object level &key content)
  (let ((format-args (format-of writer
                                level
                                (ignore-errors
                                 (slot-value object 'tag)))))
    (when (null (car format-args))
      (setf format-args (cdr format-args)))
    (apply #'format nil (car format-args)
           (mapcar (lambda (slot)
                     (or content
                         (ignore-errors
                          (slot-value object slot))
                         ""))
                   (or (cdr format-args)
                       '(content))))))

(defun format-of (writer level &optional tag)
  (let ((formats (writer-formats (writer writer))))
    (or (spec-match formats tag level)
        (case level
          (:line
           (or (spec-match formats nil :line)
               (list *default-line-format*)))
          (:inline
           (or (spec-match formats tag :line)
               (spec-match formats nil :inline)
               (spec-match formats nil :line)
               (list *default-line-format*)))
          (:stanza
           (or (spec-match formats nil :stanza)
               (list *default-stanza-format*)))
          (:document
           (or (spec-match formats nil :document)
               (list *default-document-format*)))))))

(defun struct-filters-of (writer tag level)
  (let ((filters (writer-struct-filters (writer writer))))
    (apply #'compose #'identity
           (or (spec-match filters tag level)
               (spec-match filters nil level)))))

(defun text-filters-of (writer tag level)
  (let ((filters (writer-text-filters (writer writer))))
    (apply #'compose #'identity
           (or (spec-match filters tag level)
               (spec-match filters nil level)))))

(defun line-string (line &key inlinep (writer (current-writer)))
  (let* ((tag (line-tag line))
         (level (if inlinep :inline :line))
         (line (funcall (struct-filters-of writer tag level)
                        line)))
    (funcall (text-filters-of writer tag level)
             (format-object writer line level))))

(defun stanza-string (stanza &key (writer (current-writer)))
  (let* ((tag (stanza-tag stanza))
         (level :stanza)
         ;;(format-args (format-of writer tag level))
         (stanza (funcall (struct-filters-of writer tag level)
                          stanza))
         (formatted-lines
           (loop for line in (stanza-content stanza)
                 collect
                 (line-string line
                              :inlinep (not (eq (line-tag line) tag))
                              :writer writer))))
    (funcall (text-filters-of writer tag level)
             (format-object writer stanza level :content formatted-lines))))

(defun document-string (document &key (writer (current-writer)))
  (let* ((tag nil)
         (level :document)
         ;;(format-args (format-of writer tag level))
         (formatted-stanzas
           (loop for stanza in (document-content document)
                 collect
                 (stanza-string stanza :writer writer))))
    (funcall (text-filters-of writer tag level)
             (format-object writer document level :content formatted-stanzas))))

;;; Filters

(defun fold-text (string &key (width 72) (prefix 0))
  ;; TODO: fold-text-preserving-gemini-links
  "Fold (word-wrap) STRING to WIDTH."
  (let ((words (split "\\s+" string))
        (current-width 0))
    (loop for word in words
          for word-width = (length word)
          if (>= (+ word-width current-width) width)
            do (progn (format t "~&~,,v@a " prefix word)
                      (setf current-width (+ word-width 1 prefix)))
          else
            do (progn (format t "~a " word)
                      (incf current-width (1+ word-width))))))

(defun unfold-text (string)
  "Replace newlines in STRING with spaces."
  (regex-replace-all "\\n" (chomp string) " "))

(defun combine-like (stanza)
  "Combine lines with similar tags in STANZA together."
  (let ((combined
          (reduce
           (lambda (lines line)
             (let ((this (car lines)))
               (if (and this
                        (not (link-p line))
                        (eq (line-tag this)
                            (line-tag line)))
                   (cons (make-line :tag (line-tag line)
                                    :content
                                    (format nil "~A ~A"
                                            (line-content this)
                                            (line-content line)))
                         (cdr lines))
                   (cons line lines))))
           (stanza-content stanza)
           :initial-value ())))
    (make-stanza (nreverse combined))))

;;; XXX -- buggy as heck
(defun character-wrap (string char open close)
  "Convert runs of 'CHAR ... CHAR' in STRING to 'OPEN ... CLOSE'.
CHAR must be on a word boundary.  Example:
(character-wrap \"some *body* once told me\" #\* \"<b>\" \"<\/b>\") =>
\"some <b>body</b> once told me\"."
  (regex-replace-all `(:sequence
                       (:register
                        (:alternation
                         :whitespace-char-class
                         :start-anchor
                         #\- #\[ #\( #\{ #\" #\' #\<))
                       ,char
                       (:register
                        (:sequence
                         :non-whitespace-char-class
                         (:non-greedy-repetition 0 nil :everything)
                         :non-whitespace-char-class))
                       ,char
                       (:register
                        (:alternation
                         :whitespace-char-class
                         :end-anchor
                         #\, #\. #\? #\! #\; #\: #\- #\] #\) #\}
                         #\" #\' #\>)))
                     string
                     (format nil "\\1~A\\2~A\\3" open close)))


(defun escape-html (text)
  "Escape HTML entities in TEXT."
  (ppcre:regex-replace-all
   "[&<>]" text
   (lambda (target-string start end match-start match-end reg-starts reg-ends)
     (declare (ignore start end match-end reg-starts reg-ends))
     (case (char target-string match-start)
       (#\< "&lt;")
       (#\> "&gt;")
       (#\& "&amp;")))))

(defun escape-html/pre (line)
  (if (link-p line)
      (make-link :href (link-href line)
                 :content (escape-html (link-content line)))
      (make-line :tag (line-tag line)
                 :content (escape-html (line-content line)))))

(defun add-inline-markup (line)
  "Convert *inline* =markup= to HTML in a LINE."
  (let* ((text (line-content line))
         (text (character-wrap text #\* "<b>" "</b>"))
         (text (character-wrap text #\_ "<i>" "</i>"))
         (text (character-wrap text #\= "<code>" "</code>")))
    (if (link-p line)
        (make-link :content text :href (link-href line))
        (make-line :tag (line-tag line) :content text))))

(defun verbatim-preproc (verb)
  "Give VERB the correct OPEN and CLOSE tags."
  (let ((alt (verbatim-alt verb))
        (open "<pre><code>")
        (close "</code></pre>")
        (preproc '(escape-html/pre)))

    (unless (zerop (length alt))
      (case (char alt 0)
        ((#\. #\#)                      ; .classes, #id
         (multiple-value-bind (classes id)
             (loop for x in (ppcre:all-matches-as-strings "[#.][^#.]+" alt)
                   if (eq #\. (char x 0))
                     collect (subseq x 1) into classes
                   if (eq #\# (char x 0))
                     collect (subseq x 1) into id
                   finally (return (values classes id)))
           (setf open (format nil "<pre~@?~@?><code>"
                              (if classes " class=\"~{~A~^ ~}\"" "~*")
                              classes
                              (if id " id=\"~{~A~^ ~}\"" "~*")
                              id))))

        (#\<                            ; custom HTML tag
         (if (uiop:string-prefix-p "<>" alt)
             ;; ```<> means interpret the verbatim block as raw HTML
             (setf open ""
                   close ""
                   preproc nil)
             (setf open alt
                   close (regex-replace-all "<([^\\s>]+)([^>]+)?>"
                                            open
                                            "</\\1>"))))

        (#\|                            ; pipe to an external process
         (let* ((stanza-text
                  (loop for ln in (verbatim-content verb)
                        collect (line-content ln) into ss
                        finally (return (format nil "~{~A~^~%~}" ss))))
                (command-line
                  (ppcre:split "\\s" (subseq alt 1)))
                (piped-text
                  (ignore-errors
                   (with-input-from-string (s stanza-text)
                     (uiop:run-program command-line :input s :output
                                       :string)))))
           ;; Here it's easier to just return a new stanza altogether
           (when piped-text
             (setf open (format nil "<!-- ~A -->~%" alt)
                   close "")
             (setf verb (%make-verbatim
                         :tag :verbatim
                         :content (with-input-from-string (s piped-text)
                                    (parse-lines s)))))))))

    (%make-verbatim
     :tag :verbatim
     :alt (verbatim-alt verb)
     :open open
     :close close
     :content (mapcar (apply #'compose #'identity preproc)
                      (verbatim-content verb)))))

;;; Utilities

(defun chomp (string)
  "Trim whitespace off both ends of STRING."
  (string-trim '(#\newline #\space #\tab) string))

(defun compose (function &rest more-functions) ; yoinked from alexandria
  "Returns a function composed of FUNCTION and MORE-FUNCTIONS that applies its
arguments to to each in turn, starting from the rightmost of MORE-FUNCTIONS,
and then calling the next one with the primary value of the last."
  (reduce (lambda (f g)
            (lambda (&rest arguments)
              (declare (dynamic-extent arguments))
              (funcall f (apply g arguments))))
          more-functions
          :initial-value function))

(defun alist-get (x alist &rest rest)
  (cdr (apply #'assoc x alist rest)))

(defun alist-set (x alist new &rest rest)
  (rplacd (apply #'assoc x alist rest) new))

(defsetf alist-get alist-set)
